# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import json

from doppelganger import (bayesnets, allocation)

CURRENT_VERSION = '0'


class Configuration(object):
    """A configuration to population generation.

    See the sample configuration for details on how to produce a configuration
    file.
    """

    def __init__(self,
                 household_fields,
                 person_fields,
                 household_structure,
                 person_structure,
                 preprocessing_config,
                 version
                 ):
        self.person_fields = person_fields
        self.household_fields = household_fields
        self.person_structure = person_structure
        self.household_structure = household_structure
        self.preprocessing_config = preprocessing_config
        self.version = version

    @staticmethod
    def _read_net_structure(filename):
        with open(filename) as bayes_net_file_config:
            net_json = json.loads(bayes_net_file_config.read())
            return bayesnets.define_bayes_net_structure(
                net_json['nodes'], net_json['edges']
            )

    @staticmethod
    def from_json(config_json):
        """Create a Configuration from a json blob.

        Args:
            config_json (dict): the full configuration parsed json

        Returns: Configuration to be used in population generation.
        """

        assert 'version' in config_json, "The config file is missing a 'version' field"
        assert config_json['version'] == CURRENT_VERSION, \
            "The config file version is incorrect. \
            Please upgrade your config file to version {}.".format(CURRENT_VERSION)
        network_configs = config_json['network_config_files']
        household_structure = Configuration._read_net_structure(
            network_configs['household'])
        person_structure = Configuration._read_net_structure(
            network_configs['person'])
        household_fields = tuple(config_json['household_fields'])
        person_fields = tuple(config_json['person_fields'])
        preprocessing = config_json['preprocessing']
        version = config_json['version']
        return Configuration(
            household_fields,
            person_fields,
            household_structure,
            person_structure,
            preprocessing,
            version
        )

    @staticmethod
    def from_file(infile):
        """Create a Configuration from a file.

        Args:
            infile (unicode): path to a json config file

        Returns: Configuration to be used in population generation.
        """
        with open(infile) as training_config_file:
            config_json = json.loads(training_config_file.read())
            return Configuration.from_json(config_json)

    def get_all_person_fields(self):
        """Create a tuple of combined config and the default person fields.

        Returns: Combination of allocation default person
            fields and configuration person fields as tuple
        """

        default_set = set(field.name for field in allocation.DEFAULT_PERSON_FIELDS)
        config_set = set(self.person_fields)
        return tuple(default_set.union(config_set))

    def get_all_household_fields(self):
        """Create a tuple of combined config and the default household fields.

        Returns: Combination of allocation default household
            fields and configuration household fields as tuple
        """

        default_set = set(field.name for field in allocation.DEFAULT_HOUSEHOLD_FIELDS)
        config_set = set(self.household_fields)
        return tuple(default_set.union(config_set))



# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

"""Library for reading PUMS data.

"""
from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
from builtins import str

import datetime
from enum import Enum
import math

# Constant for missing data
UNKNOWN = None


class WorkStatus(object):
    UNDER_16 = 'under16'
    NOT_WORKING = 'not_working'
    WORKING = 'working'


class EducationalAttainment(object):
    UNDER_3 = 'under3'
    NO_SCHOOL = 'no_school'
    K12 = 'k-12'
    HIGH_SCHOOL = 'high_school'
    SOME_COLLEGE = 'some_college'
    BACHELORS_DEGREE = 'bachelors_degree'
    ADVANCED_DEGREE = 'advanced_degree'


class DataCategory(Enum):
    PERSON = 1
    HOUSEHOLD = 2


class DataType(object):

    def __init__(self, name, pums_name, preprocessor, type_, possible_values):
        self.name = name
        self.pums_name = pums_name
        self.preprocessor = preprocessor
        self.type_ = type_
        self.possible_values = possible_values

    def process(self, data):
        if self.preprocessor is not None:
            return self.preprocessor(data)
        return data


def is_blank(input):
    return (
        input is UNKNOWN or
        input == '' or
        isinstance(input, float) and math.isnan(input)
    )


def generate_binning_preprocessor(bins):
    if len(bins) == 0:
        return lambda x: 'all_values'
    labels = ['<={}'.format(bins[0])]
    last_label = str(bins[0])
    for elem in bins[1:]:
        bin_label = str(elem)
        labels.append(last_label + '-' + bin_label)
        last_label = bin_label
    labels.append('{}+'.format(bins[-1]))

    def generate_bin(number):
        if is_blank(number):
            return labels[0]
        for i, bin_ in enumerate(bins):
            if int(number) <= bin_:
                return labels[i]
        return labels[-1]
    return labels, generate_bin


# Default processors


def age_discrete(age):
    if is_blank(age):
        return UNKNOWN
    age = int(age)
    if age < 18:
        return '0-17'
    if age < 35:
        return '18-34'
    if age < 65:
        return '35-64'
    return '65+'


def num_people_discrete(num_people):
    if is_blank(num_people):
        return '0'  # ACS has no equivalent to the PUMS '' value
    if int(num_people) < 4:
        return str(int(num_people))
    return '4+'


def work_status(code):
    ''' Employment status recode (PUMS 2015 code: ESR)
        b .N/A (less than 16 years old)
        1 .Civilian employed, at work
        2 .Civilian employed, with a job but not at work
        3 .Unemployed
        4 .Armed forces, at work
        5 .Armed forces, with a job but not at work
        6 .Not in labor force
    '''
    if is_blank(code):
        return WorkStatus.UNDER_16
    if code == '1' or code == '2' or code == '4' or code == '5':
        return WorkStatus.WORKING
    if code == '3' or code == '6':
        return WorkStatus.NOT_WORKING


def educational_attainment(code):
    ''' Educational attainment (PUMS 2015 code: SCHL)
    Pums_Value Category
        bb     N/A (less than 3 years old)
        01     No schooling completed
        02     Nursery school, preschool
        03     Kindergarten
        04     Grade 1
        05     Grade 2
        06     Grade 3
        07     Grade 4
        08     Grade 5
        09     Grade 6
        10     Grade 7
        11     Grade 8
        12     Grade 9
        13     Grade 10
        14     Grade 11
        15     12th grade - no diploma
        16     Regular high school diploma
        17     GED or alternative credential
        18     Some college, but less than 1 year
        19     1 or more years of college credit, no degree
        20     Associate's degree
        21     Bachelor's degree
        22     Master's degree
        23     Professional degree beyond a bachelor's degree
        24     Doctorate degree
    '''
    if is_blank(code) or code == 'bb':
        return EducationalAttainment.UNDER_3
    if code == '01':
        return EducationalAttainment.NO_SCHOOL
    if (  # Including preschool with k-12 to avoid extra category in model
            code == '02' or code == '03' or code == '04' or code == '05' or code == '06' or
            code == '07' or code == '08' or code == '09' or code == '10' or code == '11' or
            code == '12' or code == '13' or code == '14' or code == '15'
       ):
        return EducationalAttainment.K12
    if code == '16' or code == '17':
        return EducationalAttainment.HIGH_SCHOOL
    if code == '18' or code == '19' or code == '20':
        return EducationalAttainment.SOME_COLLEGE
    if code == '21':
        return EducationalAttainment.BACHELORS_DEGREE
    if code == '22' or code == '23' or code == '24':
        return EducationalAttainment.ADVANCED_DEGREE
    return UNKNOWN


def num_vehicles_discrete(num_vehicles):
    if is_blank(num_vehicles):
        return '0'  # ACS has no equivalent to the PUMS '' value
    if int(num_vehicles) < 3:
        return str(int(num_vehicles))
    return '3+'


def gender_named(input):
    if is_blank(input):
        return UNKNOWN

    input = int(input)
    if input == 1:
        return 'M'
    elif input == 2:
        return 'F'
    else:
        return UNKNOWN


def yyyy_to_age(input):
    if is_blank(input):
        return UNKNOWN
    input = str(input)
    input_year = int(input[:4])
    cur_year = datetime.date.today().year
    return cur_year - input_year


AGE = DataType('age', 'agep', age_discrete, DataCategory.PERSON,
               {'0-17', '18-34', '35-64', '65+'})

SEX = DataType('sex', 'sex', gender_named, DataCategory.PERSON, {'M', 'F'})

WORKING = DataType('working', 'esr', work_status, DataCategory.PERSON, {'under-16', 'Y', 'N'})

EDUCATION = DataType('education', 'schl', educational_attainment, DataCategory.PERSON,
                     {'under-3', 'none', 'k-12', 'high-school', 'some-college', 'bachelors-degree',
                      'advanced-degree'})

_income_labels, _income_preprocessor = generate_binning_preprocessor([40000])
INDIVIDUAL_INCOME = DataType(
    'individual_income', 'wagp', _income_preprocessor, DataCategory.PERSON, _income_labels
)

HOUSEHOLD_INCOME = DataType(
    'household_income', 'fincp', _income_preprocessor, DataCategory.HOUSEHOLD, _income_labels
)

NUM_VEHICLES = DataType(
    'num_vehicles', 'veh', num_vehicles_discrete, DataCategory.HOUSEHOLD,
    {'0', '1', '2', '3+'}
)

NUM_PEOPLE = DataType(
    'num_people', 'np', num_people_discrete, DataCategory.HOUSEHOLD,
    {'1', '2', '3', '4+'}
)

HOUSEHOLD_WEIGHT = DataType(
    'household_weight', 'wgtp', int, DataCategory.HOUSEHOLD, None
)

PERSON_WEIGHT = DataType(
    'person_weight', 'pwgtp', int, DataCategory.PERSON, None
)

SERIAL_NUMBER = DataType('serial_number', 'serialno', None, None, None)

PUMA = DataType('puma', 'puma', str, None, None)

STATE = DataType('state', 'st', str, None, None)

TRAFFIC_ANALYSIS_ZONE = DataType('taz', None, str, None, None)

PUMS_INPUTS = [
    AGE,
    SEX,
    WORKING,
    EDUCATION,
    INDIVIDUAL_INCOME,
    HOUSEHOLD_INCOME,
    NUM_VEHICLES,
    NUM_PEOPLE,
    HOUSEHOLD_WEIGHT,
    PERSON_WEIGHT,
    SERIAL_NUMBER,
    STATE,
    PUMA
]
NAME_TO_DATATYPE = {datatype.name: datatype for datatype in PUMS_INPUTS}

# Non-PUMS data types
TRACT = DataType('tract', None, None, None, None)
COUNT = DataType('count', None, None, None, None)
REPEAT_INDEX = DataType('repeat_index', None, None, None, None)
HOUSEHOLD_ID = DataType('household_id', None, None, None, None)



# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import pandas

from doppelganger import inputs


class DataSource(object):

    @classmethod
    def from_csv(cls, infile):
        raise NotImplementedError()

    def write(self, outfile):
        self.data.to_csv(outfile)


class DirtyDataSource(DataSource):

    def __init__(self, data, name_map):
        self.data = data
        self.name_map = name_map

    def clean(self, field_names, preprocessor, state=None, puma=None):
        cleaned_data = preprocessor.process_dataframe(self.data, field_names, self.name_map)
        if puma is not None:
            cleaned_data = cleaned_data[
                    (cleaned_data[inputs.STATE.name].astype(str) == str(state)) &
                    (cleaned_data[inputs.PUMA.name].astype(str) == str(puma))
                ]
        return CleanedData(cleaned_data)


class PumsData(DirtyDataSource):

    HOUSEHOLD_TABLE = 'households'
    PERSONS_TABLE = 'persons'

    def __init__(self, data):
        name_map = {field.name: field.pums_name for field in inputs.PUMS_INPUTS}
        return super(PumsData, self).__init__(data, name_map)

    @staticmethod
    def from_database(conn, state_id, puma_id, schema_name, table_name, fields):
        columns = ', '.join(field.pums_name for field in fields)
        query = ('''
            SELECT {columns}
            FROM {schema}.{table}
            WHERE ST=\'{state_id}\' AND PUMA=\'{puma_id}\'
            ORDER BY SERIALNO
            ;''').format(
                columns=columns,
                schema=schema_name,
                table=table_name,
                state_id=state_id,
                puma_id=puma_id
            )
        return PumsData(pandas.read_sql_query(query, conn))

    @staticmethod
    def from_csv(infile):
        data = pandas.read_csv(infile)
        return PumsData(data)


class CleanedData(DataSource):

    def __init__(self, data):
        self.data = data

    @staticmethod
    def from_csv(infile):
        data = pandas.read_csv(infile)
        return CleanedData(data)



from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import unittest
import mock
import os
from doppelganger.scripts.fetch_pums_data_from_db import link_fields_to_inputs, fetch_pums_data
from doppelganger import inputs, config, allocation


class TestFetchPumsDataFromDB(unittest.TestCase):

    def _mock_config(self):
        conf = {
            "person_fields": [
                "age",
                "sex",
                "individual_income"
            ],
            "household_fields": [
                "num_people",
                "household_income",
                "num_vehicles"
            ],
            "preprocessing": {
                "individual_income": {
                    "bins": [
                        0,
                        20000,
                        40000,
                        80000,
                        100000
                    ]
                }
            },
            "network_config_files": {
                "person": os.path.dirname(__file__)
                + "/../examples/sample_data/sample_person_bn.json",
                "household": os.path.dirname(__file__)
                + "/../examples/sample_data/sample_household_bn.json"
            },
            "version": "0"
        }
        return config.Configuration.from_json(conf)

    _mock_params = {
                    'state_id': '01',
                    'puma_id': '00001',
                    'output_dir': '.',
                    'db_host': 'host1',
                    'db_database': 'database1',
                    'db_schema': 'schema1',
                    'db_table': 'persons',  # Hard-Coded for now
                    'db_user': 'user1',
                    'db_password': 'password1',
                    'census_api_key': 'census_key1',
                    'puma_tract_mappings': 'puma_tract_mappings',
                }

    def _mock_legit_person_input_list(self):
        return ['age', 'sex', 'individual_income']

    def _mock_illegit_person_input_list(self):
        return ['not_a_field', 'sex', 'individual_income']

    def test_link_fields_to_inputs(self):
        self.assertEqual(set([inputs.AGE, inputs.SEX, inputs.INDIVIDUAL_INCOME]),
                         link_fields_to_inputs(self._mock_legit_person_input_list()))
        self.assertRaises(ValueError, link_fields_to_inputs, self._mock_illegit_person_input_list())

    @mock.patch('doppelganger.scripts.fetch_pums_data_from_db.psycopg2')
    @mock.patch('doppelganger.scripts.fetch_pums_data_from_db.datasource.PumsData')
    def test_fetch_pums_data(self, mock_PumsData, mock_psycopg2):
        mock_config = self._mock_config()
        mock_db_connection = mock.Mock()
        mock_psycopg2.connect.return_value = mock_db_connection

        default_person_fields = allocation.DEFAULT_PERSON_FIELDS

        # local_person_fields = link_fields_to_inputs(mock_config.person_fields)
        # person_fields = default_person_fields.union(local_person_fields)

        with mock.patch(
                'doppelganger.scripts.fetch_pums_data_from_db.allocation'
                ) as mock_allocation:
            mock_allocation.DEFAULT_PERSON_FIELDS = \
                mock.PropertyMock(default_person_fields)
            persons_data, households_data = fetch_pums_data(
                            state_id=self._mock_params['state_id'],
                            puma_id=self._mock_params['puma_id'],
                            configuration=mock_config,
                            db_host=self._mock_params['db_host'],
                            db_database=self._mock_params['db_database'],
                            db_schema=self._mock_params['db_schema'],
                            db_user=self._mock_params['db_user'],
                            db_password=self._mock_params['db_password']
                        )
        mock_psycopg2.connect.assert_called()
        mock_PumsData.from_database.assert_called()



# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import logging
import cvxpy as cvx
import numpy as np

logging.basicConfig(filename='logs', filemode='a', level=logging.INFO)


def _insert_append(arr, indices, values, axis=0):
    """Insert / Append values to array along given axis

    Args:
    arr (numpy array): Array to insert/append values
    indices (list(int)): indices before which values are inserted
    values (numpy array): Values to insert into arr
    axis (int): Axis along which to insert values

    Returns:
    numpy array: A copy of arr with values inserted and appended
    """
    insert_filter = indices < arr.shape[axis]
    insert_inds = indices[insert_filter]
    n_append = indices[~insert_filter].shape[0]
    append_shape = (1, n_append) if axis else (n_append, 1)
    arr_append = np.tile(values, append_shape)

    arr_update = np.insert(arr, insert_inds, values, axis=axis)
    arr_update = np.concatenate((arr_update, arr_append), axis=axis)
    return arr_update


def balance_cvx(hh_table, A, w, mu=None, verbose_solver=False):
    """Maximum Entropy allocaion method for a single unit

    Args:
        hh_table (numpy matrix): Table of households categorical data
        A (numpy matrix): Area marginals (controls)
        w (numpy array): Initial household allocation weights
        mu (numpy array): Importance weights of marginals fit accuracy
        verbose_solver (boolean): Provide detailed solver info

    Returns:
        (numpy matrix, numpy matrix): Household weights, relaxation factors
    """

    n_samples, n_controls = hh_table.shape
    x = cvx.Variable(n_samples)

    if mu is None:
        objective = cvx.Maximize(
            cvx.sum_entries(cvx.entr(x) + cvx.mul_elemwise(cvx.log(w.T), x))
        )

        constraints = [
            x >= 0,
            x.T * hh_table == A,
        ]
        prob = cvx.Problem(objective, constraints)
        prob.solve(solver=cvx.SCS, verbose=verbose_solver)

        return x.value

    else:
        # With relaxation factors
        z = cvx.Variable(n_controls)

        objective = cvx.Maximize(
            cvx.sum_entries(cvx.entr(x) + cvx.mul_elemwise(cvx.log(w.T), x)) +
            cvx.sum_entries(mu * (cvx.entr(z)))
        )

        constraints = [
            x >= 0,
            z >= 0,
            x.T * hh_table == cvx.mul_elemwise(A, z.T),
        ]
        prob = cvx.Problem(objective, constraints)
        prob.solve(solver=cvx.SCS, verbose=verbose_solver)

        return x.value, z.value


def balance_multi_cvx(hh_table, A, B, w, mu=1000., meta_mu=1000., verbose_solver=False):
    """Maximum Entropy allocation method for multiple balanced units

    Args:
        hh_table (numpy matrix): Table of households categorical data
        A (numpy matrix): Area marginals (controls)
        B (numpy matrix): Meta-marginals
        w (numpy array): Initial household allocation weights
        mu (float): Importance weights of marginals for accuracy of fit
        meta_mu (float): Importance weights of meta-marginals for accuracy of fit
        verbose_solver (boolean): Provide detailed solver info

    Returns:
        (numpy matrix, numpy matrix, numpy matrix): Household weights,
            relaxation factors, relaxation factors,
    """

    n_samples, n_controls = hh_table.shape

    # Solver won't converge with zero marginals. Identify and remove.
    zero_marginals = np.where(~A.any(axis=1))[0]
    zero_weights = np.zeros((1, n_samples))

    if zero_marginals.size:
        logging.info(
            '%i tract(s) with zero marginals encountered. '
            'Setting weights to zero'.format(zero_marginals.size)
        )

        # Need to remove problem tracts and add a row of zeros later
        A = np.delete(A, zero_marginals, axis=0)
        w = np.delete(w, zero_marginals, axis=0)
        mu = np.delete(mu, zero_marginals, axis=1)

    n_tracts = w.shape[0]
    x = cvx.Variable(n_tracts, n_samples)

    # Relative weights of tracts
    # (need to reshape for numpy broadcasting)
    wa = (np.sum(A, axis=1) / np.sum(A)).reshape(-1, 1)
    w_relative = (np.array(w) * np.array(wa))

    # With relaxation factors
    z = cvx.Variable(n_controls, n_tracts)
    q = cvx.Variable(n_controls)

    identity = np.ones((n_tracts, 1))

    solved = False
    while not solved:
        objective = cvx.Maximize(
            cvx.sum_entries(
                cvx.entr(x) + cvx.mul_elemwise(cvx.log(np.e * w_relative), x)
            ) +
            cvx.sum_entries(
                cvx.mul_elemwise(
                    mu, cvx.entr(z) + cvx.mul_elemwise(cvx.log(np.e), z)
                )
            ) +
            cvx.sum_entries(
                cvx.mul_elemwise(
                    meta_mu, cvx.entr(q) + cvx.mul_elemwise(cvx.log(np.e), q)
                )
            )
        )

        constraints = [
            x >= 0,
            z >= 0,
            q >= 0,
            x * hh_table == cvx.mul_elemwise(A, z.T),
            cvx.mul_elemwise(A.T, z) * identity == cvx.mul_elemwise(B.T, q)
        ]

        prob = cvx.Problem(objective, constraints)

        try:
            prob.solve(verbose=verbose_solver)
            solved = True

        except cvx.SolverError:
            if np.all(mu == 1):
                # We can't reduce mu any further
                break
            mu = np.where(mu > 10, mu - 10, 1)
            logging.info('Solver error encountered. Importance weights have been relaxed.')

    if not np.any(x.value):
        logging.exception('Solution infeasible. Using initial weights.')

    # If we didn't get a value return the initial weights
    weights_out = x.value if np.any(x.value) else w_relative

    # Insert zeros
    if zero_marginals.size:
        # Due to numpy insert behavior, we need to differentiate between the
        # values that go into the middle of the array, and the values that get appended
        weights_out = _insert_append(
            weights_out, zero_marginals, zero_weights, axis=0)

    return weights_out


def discretize_multi_weights(hh_table, x, gamma=100., verbose_solver=False):
    """Discretize weights in household table for multiple tracts

    Arguments:
        hh_table (numpy matrix): Table of households categorical data
        x (numpy matrix): Household weights
        gamma (float): Relaxation weight
        verbose_solver (boolean): Provide detailed solver info

    Returns:
        numpy array: Discretized household weights
    """

    n_samples, n_controls = hh_table.shape

    # Solver won't converge with zero weights. Identify and remove.
    zero_weights_inds = np.where(~x.any(axis=1))[0]
    zero_weights = np.zeros((1, n_samples))

    if zero_weights_inds.size:
        logging.info(
            '{} tract(s) with zero weight rows encountered. '
            'Setting weights to zero'.format(zero_weights_inds.size)
        )

        # Need to remove problem tracts and add a row of zeros later
        x = np.delete(x, zero_weights_inds, axis=0)

    n_tracts = x.shape[0]

    # Integerize x values
    x_int = x.astype(int)

    # Get residuals in new marginals from truncating to int
    A_residuals = np.dot(x, hh_table) - np.dot(x_int, hh_table)
    x_residuals = x - x_int

    # Coefficients in objective function
    x_log = np.log(x_residuals)

    # Decision variables for optimization
    y = cvx.Variable(n_tracts, n_samples)

    # Relaxation factors
    U = cvx.Variable(n_tracts, n_controls)
    V = cvx.Variable(n_tracts, n_controls)

    objective = cvx.Maximize(
        cvx.sum_entries(
            cvx.sum_entries(cvx.mul_elemwise(x_log, y), axis=1) -
            (gamma) * cvx.sum_entries(U, axis=1) -
            (gamma) * cvx.sum_entries(V, axis=1)
        )
    )

    constraints = [
        y * hh_table <= A_residuals + U,
        y * hh_table >= A_residuals - V,
        U >= 0,
        V >= 0,
        y >= 0,
        y <= 1.0
    ]

    prob = cvx.Problem(objective, constraints)

    try:
        prob.solve(verbose=verbose_solver)

    except cvx.SolverError:
        logging.exception(
            'Solver error encountered in weight discretization. Weights will be rounded.')

    weights_out = y.value if np.any(y.value) else x_residuals

    # Insert zeros
    if zero_weights_inds.size:
        weights_out = _insert_append(weights_out, zero_weights_inds, zero_weights, axis=0)

    # Make results binary and return
    return np.array(weights_out > 0.5).astype(int)



# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from mock import MagicMock, patch

import unittest
import pandas

from doppelganger import inputs, Population, HouseholdAllocator


class TestPopulationGen(unittest.TestCase):

    def _mock_allocated(self):
        def mock_person(serialno, age, sex, income):
            return {
                'serial_number': serialno,
                'age': age,
                'sex': sex,
                'individual_income': income
            }
        allocated_persons = pandas.DataFrame([
            mock_person('b', '35-64', 'F', '40k+'),
            mock_person('b', '35-64', 'M', 'None'),
        ])

        def mock_household(serialno, num_people, num_vehicles, income, count, tract):
            return {
                'serial_number': serialno,
                'num_people': num_people,
                'num_vehicles': num_vehicles,
                'household_income': income,
                'count': count,
                'tract': tract,
            }
        allocated_households = pandas.DataFrame([
            mock_household('b', '6+', '2', '40k+', count=2, tract='tract1'),
            mock_household('b', '6+', '2', '40k+', count=2, tract='tract2'),
        ])
        return HouseholdAllocator(allocated_households, allocated_persons)

    def _mock_model(self, fields, generated):
        model = MagicMock()
        model.fields = fields
        model.segmenter = MagicMock(return_value='one_bucket')
        model.generate = MagicMock(
            return_value=generated)
        return model

    def _check_household_output(self, dataframe):
        self.assertSequenceEqual(
            dataframe[inputs.TRACT.name].tolist(), ('tract1', 'tract1', 'tract2', 'tract2'))
        self.assertSequenceEqual(
            dataframe[inputs.SERIAL_NUMBER.name].tolist(), ('b', 'b', 'b', 'b'))
        self.assertSequenceEqual(dataframe[inputs.REPEAT_INDEX.name].tolist(), (0, 1, 0, 1))
        self.assertEqual(dataframe[inputs.HOUSEHOLD_ID.name].tolist()[0], 'tract1-b-0')

    def _check_person_output(self, dataframe):
        self.assertSequenceEqual(
            dataframe[inputs.TRACT.name].tolist(), ('tract1', 'tract1', 'tract2', 'tract2',
                                                    'tract1', 'tract1', 'tract2', 'tract2'))
        self.assertSequenceEqual(
            dataframe[inputs.REPEAT_INDEX.name].tolist(), (0, 1, 0, 1, 0, 1, 0, 1))
        self.assertSequenceEqual(dataframe[inputs.AGE.name].tolist(), ['35-64'] * 8)
        self.assertEqual(dataframe[inputs.HOUSEHOLD_ID.name].tolist()[0], 'tract1-b-0')

    def test_generate_persons_simple(self):
        person_model = self._mock_model(
            [inputs.AGE.name, inputs.SEX.name],
            # Returns two people regardles of count passed in
            generated=[('35-64', 'F'), ('35-64', 'F')]
        )
        allocations = self._mock_allocated()
        population = Population.generate(
            allocations, person_model, MagicMock())

        evidence = ((inputs.AGE.name, '35-64'), (inputs.SEX.name, 'M'))

        person_model.generate.assert_called_with(
            'one_bucket', evidence, count=2)
        self._check_person_output(population.generated_people)

    def test_generate_households_simple(self):
        household_model = self._mock_model(
            [inputs.NUM_PEOPLE.name],
            generated=[('6+',), ('6+',)]
        )
        allocations = self._mock_allocated()
        population = Population.generate(
            allocations, MagicMock(), household_model)

        evidence = ((inputs.NUM_PEOPLE.name, '6+'),)

        household_model.generate.assert_called_with(
            'one_bucket', evidence, count=2)

        self.assertIn(inputs.NUM_PEOPLE.name, population.generated_households)
        self._check_household_output(population.generated_households)

    def test_read_from_file(self):
        read_csv = MagicMock(return_value=pandas.DataFrame())
        with patch('pandas.read_csv', read_csv):
            population = Population.from_csvs('persons_file', 'households_file')
        assert type(population) == Population
        read_csv.assert_any_call('households_file')
        read_csv.assert_any_call('persons_file')

    def test_write_to_file(self):
        persons = MagicMock()
        households = MagicMock()
        population = Population(persons, households)

        population.write(persons_outfile='persons_file', households_outfile='households_file')
        persons.to_csv.assert_called_once_with('persons_file')
        households.to_csv.assert_called_once_with('households_file')



# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import unittest
import math
import sys

import pandas
from mock import patch, mock_open
import numpy

from doppelganger import (
    bayesnets,
    inputs,
    datasource,
    BayesianNetworkModel,
    Preprocessor
)


class BayesNetTests(unittest.TestCase):

    def _one_person_house(self):
        return inputs.num_people_discrete(num_people=1)

    def _two_person_house(self):
        return inputs.num_people_discrete(num_people=2)

    def _mock_household_input(self):
        def mock_household(serialno, num_people, num_vehicles, income):
            return {
                'serial_number': serialno,
                'num_people': num_people,
                'num_vehicles': num_vehicles,
                'household_income': income
            }
        return datasource.CleanedData(pandas.DataFrame([
            mock_household('a', '1', '1', '<=0'),
            mock_household('b', '2', '6+', '40k+'),
            mock_household('c', '1', 'None', '<=0')
        ]))

    def _mock_person(self, serialno, age, sex, income, weight=1):
        return {
            'serial_number': serialno,
            'age': age,
            'sex': sex,
            'individual_income': income,
            'person_weight': weight
        }

    def _mock_people_input(self, weight=1):
        return datasource.CleanedData(pandas.DataFrame([
            self._mock_person('a', '0-17', 'M', '<=0', weight),
            self._mock_person('b', '35-64', 'F', '40k+', weight),
            self._mock_person('b', '65+', 'M', '0-40k', weight),
            self._mock_person('c', '18-34', 'M', '<=0', weight)
        ]))

    def _mock_nodes(self):
        return ('age', 'sex', 'income')

    def _mock_edges(self):
        return {
            'age': ['sex', 'income'],
            'sex': ['income']
        }

    def _household_fields(self):
        return [inputs.HOUSEHOLD_INCOME.name, inputs.NUM_VEHICLES.name]

    def _person_fields(self):
        return [
            inputs.AGE.name,
            inputs.SEX.name,
            inputs.INDIVIDUAL_INCOME.name
        ]

    def _mock_persons_missing(self):
        return datasource.CleanedData(pandas.DataFrame([
            self._mock_person('d', '0-17', 'M', inputs.UNKNOWN),
            self._mock_person('e', '35-64', 'F', '40k+'),
            self._mock_person('e', inputs.UNKNOWN, 'M', '0-40k'),
            self._mock_person('f', '18-34', 'M', '<=0')
        ]))

    def _person_structure(self):
        return ((), (0,), (0, 1))

    def _household_structure(self):
        return ((), (0,))

    def _person_segmenter(self):
        serialno_to_household = {
            'a': self._one_person_house(),
            'b': self._two_person_house(),
            'c': self._one_person_house(),
            'd': self._one_person_house(),
            'e': self._two_person_house(),
            'f': self._one_person_house()
        }

        return lambda x: serialno_to_household[x['serial_number']]

    def _household_segmenter(self):
        return lambda x: x['num_people']

    def _mock_household_collection(self):
        household_data = self._mock_household_input()
        people_data = self._mock_people_input()
        people_training_data = bayesnets.SegmentedData.from_data(
            people_data, self._person_fields(), 'person_weight', self._person_segmenter()
        )
        household_training_data = bayesnets.SegmentedData.from_data(
            household_data, self._household_fields(), segmenter=self._household_segmenter()
        )
        household_model = bayesnets.BayesianNetworkModel.train(
            household_training_data, self._household_structure(), self._household_fields()
        )
        person_model = bayesnets.BayesianNetworkModel.train(
            people_training_data, self._person_structure(), self._person_fields()
        )
        return household_model, person_model

    def test_read_households(self):
        household_data = self._mock_household_input()
        training_data = bayesnets.SegmentedData.from_data(
            household_data, self._household_fields(), segmenter=self._household_segmenter()
        )
        self.assertEqual(training_data.num_rows_data(), 3)
        expected_types = set([self._one_person_house(), self._two_person_house()])
        self.assertSetEqual(expected_types, set(training_data.types()))

    def test_read_people(self):
        people_data = self._mock_people_input()
        training_data = bayesnets.SegmentedData.from_data(
            people_data, self._person_fields(), 'person_weight', self._person_segmenter()
        )

        self.assertEqual(len(training_data.type_to_data[self._one_person_house()]), 2)
        self.assertEqual(len(training_data.type_to_data[self._two_person_house()]), 2)

    def test_read_people_weighted(self):
        people_data = self._mock_people_input(weight=3)
        training_data = bayesnets.SegmentedData.from_data(
            people_data, self._person_fields(), 'person_weight', self._person_segmenter()
        )
        self.assertEqual(training_data.num_rows_data(), 12)

    def test_generate_person(self):
        _, person_model = self._mock_household_collection()

        # Check that values are properly fixed based on
        # observation
        person = person_model.generate(self._two_person_house(), ((str('age'), str('65+')),),)[0]
        age_index = self._person_fields().index(inputs.AGE.name)
        self.assertEqual(person[age_index], '65+')

        person = person_model.generate(self._two_person_house(), ((str('sex'), str('F')),),)[0]
        sex_index = self._person_fields().index(inputs.SEX.name)
        self.assertEqual(person[sex_index], 'F')

    def test_generate_person_no_data(self):
        _, person_model = self._mock_household_collection()
        with self.assertRaises(ValueError):
            # We never observed any sex '2' in single-person
            # households, so we should not be able to generate it
            person = person_model.generate(self._one_person_house(), ((str('sex'), str('F')),),)
            sex_index = self._person_fields().index(inputs.SEX.name)
            self.assertEqual(person[sex_index], 'F')

    def test_generate_person_multiple(self):
        _, person_model = self._mock_household_collection()
        people = person_model.generate(
            self._two_person_house(), ((str('age'), str('65+')),), count=5
        )
        self.assertEqual(len(people), 5)
        age_index = self._person_fields().index(inputs.AGE.name)
        for person in people:
            self.assertEqual(person[age_index], '65+')

    def _check_household_generate(self, household_model):
        household = household_model.generate(
            self._two_person_house(), ((inputs.HOUSEHOLD_INCOME.name, str('40k+')),))[0]
        vehicle_count_index = self._household_fields().index(inputs.NUM_VEHICLES.name)
        self.assertEqual(household[vehicle_count_index], '6+')

    def test_generate_household(self):
        household_model, _ = self._mock_household_collection()

        # Check that values are properly fixed based on
        # observation
        self._check_household_generate(household_model)

    def test_load_bayes_net(self):
        nodes = self._mock_nodes()
        edges = self._mock_edges()
        structure = bayesnets.define_bayes_net_structure(nodes, edges)
        # pomegranate requires the use of tuples for edges but edges aren't ordered
        self.assertSequenceEqual(
            [set(node) for node in structure],
            [set(), set((0,)), set((0, 1))],
        )

    def test_train_bayes_net_from_generated_structure(self):
        '''Ensure the structures created by define_bayes_net_structure
        are compatible with pomegranate
        '''
        nodes = self._mock_nodes()
        edges = self._mock_edges()
        person_structure = bayesnets.define_bayes_net_structure(nodes, edges)
        person_training_data = self._mock_people_input()
        training_data = bayesnets.SegmentedData.from_data(
            person_training_data, self._person_fields(), segmenter=self._person_segmenter()
        )
        self.assertEquals(
            type(
                BayesianNetworkModel.train(
                    training_data,
                    person_structure,
                    self._person_fields,
                    prior_data=None
                )
            ),
            bayesnets.BayesianNetworkModel
        )

    def test_to_from_json(self):
        household_model, _ = self._mock_household_collection()
        household_string = household_model.to_json()
        household_model_new = BayesianNetworkModel.from_json(household_string)
        self.assertSequenceEqual(household_model.fields, household_model_new.fields)
        self._check_household_generate(household_model_new)

    def test_update_missing(self):
        _, person_model = self._mock_household_collection()
        missing_data = self._mock_persons_missing()
        training_data = bayesnets.SegmentedData.from_data(
            missing_data, self._person_fields(), segmenter=self._person_segmenter()
        )
        person_model.update(training_data)
        person = person_model.generate(self._two_person_house(), ((str('age'), str('65+')),),)[0]
        age_index = self._person_fields().index(inputs.AGE.name)
        self.assertEqual(person[age_index], '65+')

    def test_update_missing_iterations_inertia(self):
        _, person_model = self._mock_household_collection()
        missing_data = self._mock_persons_missing()
        training_data = bayesnets.SegmentedData.from_data(
            missing_data, self._person_fields(), segmenter=self._person_segmenter(),
        )
        person_model.update(training_data, max_iterations=5, inertia=.5)
        person = person_model.generate(self._two_person_house(), ((str('age'), str('65+')),),)[0]
        age_index = self._person_fields().index(inputs.AGE.name)
        self.assertEqual(person[age_index], '65+')

    def test_prior_creation(self):
        all_values = bayesnets.generate_laplace_prior_data(
            (inputs.AGE.name, inputs.SEX.name), Preprocessor())
        expected = {
            ('0-17', 'M'),
            ('18-34', 'M'),
            ('35-64', 'M'),
            ('65+', 'M'),
            ('0-17', 'F'),
            ('18-34', 'F'),
            ('35-64', 'F'),
            ('65+', 'F'),
        }
        self.assertSetEqual(expected, all_values)

    def test_generate_with_prior(self):
        network = BayesianNetworkModel.train(
            bayesnets.SegmentedData({'one_bucket': [('35-64', 'F', '40k+')]}),
            self._person_structure(),
            self._person_fields(),
            prior_data={('35-64', 'F', '40k+')}
        )
        person = network.generate('one_bucket', ())[0]
        self.assertEquals(person, ('35-64', 'F', '40k+'))

    def test_generate_with_prior_non_existing(self):
        network = BayesianNetworkModel.train(
            bayesnets.SegmentedData({'one_bucket': [('65+', 'M', '40k+')]}),
            self._person_structure(),
            self._person_fields(),
            prior_data={('35-64', 'F', '40k+')}
        )
        person = network.generate('one_bucket', ((str('sex'), str('F')),))[0]
        self.assertEquals(person, ('35-64', 'F', '40k+'))

        person = network.generate('one_bucket', ((str('sex'), str('M')),))[0]
        self.assertEquals(person, ('65+', 'M', '40k+'))

    def test_evaluate_network(self):
        people_data = self._mock_people_input()
        people_training_data = bayesnets.SegmentedData.from_data(
            people_data, self._person_fields(), 'person_weight',
            self._person_segmenter()
        )
        person_model = bayesnets.BayesianNetworkModel.train(
            people_training_data, self._person_structure(),
            self._person_fields()
        )
        likelihoods = person_model.log_likelihood(people_training_data)
        one_person = math.exp(likelihoods[self._one_person_house()])
        self.assertAlmostEqual(one_person, .25)
        two_person = math.exp(likelihoods[self._two_person_house()])
        self.assertAlmostEqual(two_person, .25)

    def test_evaluate_zero_likelihood(self):
        people_data = self._mock_people_input()
        people_training_data = bayesnets.SegmentedData.from_data(
            people_data, self._person_fields(), 'person_weight', self._person_segmenter()
        )
        person_model = bayesnets.BayesianNetworkModel.train(
            people_training_data, self._person_structure(), self._person_fields()
        )
        zero_prob_data = bayesnets.SegmentedData({self._one_person_house(): [('0-17', 'F', '<=0')]})
        likelihoods = person_model.log_likelihood(zero_prob_data)
        one_person = math.exp(likelihoods[self._one_person_house()])
        self.assertAlmostEqual(one_person, 0)

    def test_generate_dataframes(self):
        _, person_model = self._mock_household_collection()
        dataframes = person_model.probabilities_as_dataframes()
        self.assertSetEqual(set(dataframes.keys()), {'1', '2'})

        def _check_dataframe(dataframe, expected_cols,
                             expected_contents, expected_rows=None):
            self.assertSetEqual(set(dataframe), expected_cols)
            if expected_rows is not None:
                # Doesn't need to be checked for series
                self.assertSetEqual(set(dataframe.index), expected_rows)
            numpy.testing.assert_array_equal(dataframe, expected_contents)

        # Check single-person state 0
        expected_columns = {'0-17', '18-34'}
        _check_dataframe(dataframes['1'][0], expected_columns, [[.5, .5]])

        # Check single-person state 2
        expected_rows = {('0-17', 'M'), ('18-34', 'M',)}
        expected_columns = {'<=0'}
        _check_dataframe(
            dataframes['1'][2], expected_columns, [[1.], [1.]], expected_rows
        )

    def test_read_write(self):
        household_model, _ = self._mock_household_collection()

        def _check_network(household_model_new):
            self.assertSequenceEqual(household_model.fields, household_model_new.fields)
            self._check_household_generate(household_model_new)

        builtin_module_name = 'builtins' if sys.version_info.major == 3 else '__builtin__'
        with patch('{}.open'.format(builtin_module_name), new_callable=mock_open()) as open_mock:
            household_model.write('file')
            open_mock.assert_called_once_with('file', 'w')
            # Check that correct json is written
            json_net = open_mock.return_value.__enter__.return_value.write.call_args[0][0]
            household_model_new = BayesianNetworkModel.from_json(
                json_net, self._household_segmenter())
            _check_network(household_model_new)

            # Check that json is correctly read
            open_mock.return_value.__enter__.return_value.read.return_value = json_net
            household_model_new = BayesianNetworkModel.from_file(
                'file', self._household_segmenter())
            _check_network(household_model_new)



# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import unittest


import os.path
from doppelganger import (
    allocation,
    Configuration,
)

CURRENT_VERSION = '0'


class TestConfig(unittest.TestCase):

    def _mock_config_files(self):
        return {
            "person_fields": [
                "age",
                "sex",
                "individual_income"
            ],
            "household_fields": [
                "num_people",
                "household_income",
                "num_vehicles"
            ],
            "preprocessing": {
                "individual_income": {
                    "bins": [
                        0,
                        20000,
                        40000,
                        80000,
                        100000
                    ]
                }
            },
            "network_config_files": {
                "person": os.path.dirname(__file__)
                + "/../examples/sample_data/sample_person_bn.json",
                "household": os.path.dirname(__file__)
                + "/../examples/sample_data/sample_household_bn.json"
            },
            "version": "0"
        }

    def _mock_household_bn(self):
        return {
            "type": "household",
            "nodes": [
                "num_people",
                "household_income",
                "num_vehicles"
            ],
            "edges": {
                "num_people": [
                    "household_income",
                    "num_vehicles"
                ],
                "household_income": [
                    "num_vehicles"
                ]
            }
        }

    def _mock_person_bn(self):
        return {
            "type": "person",
            "nodes": [
                "age",
                "sex",
                "income"
            ],
            "edges": {
                "age": [
                    "income"
                ],
                "sex": [
                    "income"
                ]
            }
        }

    def test_from_json_version_empty(self):
        config_json = self._mock_config_files()
        del config_json['version']
        self.assertRaises(AssertionError, Configuration.from_json, config_json)

    def test_from_json_version_wrong(self):
        config_json = self._mock_config_files()
        config_json['version'] = '1'
        self.assertRaises(AssertionError, Configuration.from_json, config_json)

    def test_from_json_version_correct(self):
        config_json = self._mock_config_files()
        config = Configuration.from_json(config_json)
        self.assertEqual(config.version, CURRENT_VERSION)

    def test_get_all_person_fields(self):
        config_json = self._mock_config_files()
        config = Configuration.from_json(config_json)
        combined_set = set(config.get_all_person_fields())
        correct_set = set(tuple(set(
            field.name for field in allocation.DEFAULT_PERSON_FIELDS).union(
                set(config.person_fields))))
        self.assertEqual(combined_set, correct_set)

    def test_get_all_household_fields(self):
        config_json = self._mock_config_files()
        config = Configuration.from_json(config_json)
        combined_set = set(config.get_all_household_fields())
        correct_set = set(tuple(set(
            field.name for field in allocation.DEFAULT_HOUSEHOLD_FIELDS).union(
                set(config.household_fields))))
        self.assertEqual(combined_set, correct_set)



# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import unittest

from doppelganger import Preprocessor


class PreprocessingTest(unittest.TestCase):

    def test_binning_generator(self):
        bins = [0, 20000, 40000, 60000]
        config = {
            'test_input': {
                'bins': bins
            }
        }
        preprocessor = Preprocessor.from_config(config)
        preprocess = preprocessor.input_to_preprocessor['test_input']
        self.assertEqual(preprocess(0), '<=0')
        self.assertEqual(preprocess(10000), '0-20000')
        self.assertEqual(preprocess(100000), '60000+')



# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import logging
import pandas
import requests

logging.basicConfig(filename='logs', filemode='a', level=logging.INFO)


CONTROLS = {
    'num_people': {
        'count': [
            'B11016_010E',
            'B11016_003E', 'B11016_011E',
            'B11016_012E', 'B11016_004E',
            'B11016_005E', 'B11016_006E', 'B11016_007E', 'B11016_008E',
            'B11016_013E', 'B11016_014E', 'B11016_015E', 'B11016_016E'
        ],
        '1': ['B11016_010E'],
        '2': ['B11016_003E', 'B11016_011E'],
        '3': ['B11016_012E', 'B11016_004E'],
        '4+': [
            'B11016_005E', 'B11016_006E', 'B11016_007E', 'B11016_008E',
            'B11016_013E', 'B11016_014E', 'B11016_015E', 'B11016_016E'
        ],
    },
    'age': {
        # ages 0 - 17
        '0-17': [
            'B01001_003E', 'B01001_004E', 'B01001_005E', 'B01001_006E',
            'B01001_027E', 'B01001_028E', 'B01001_029E', 'B01001_030E'
        ],
        # ages 18 - 34
        '18-34': [
            'B01001_007E', 'B01001_008E', 'B01001_009E', 'B01001_010E',
            'B01001_011E', 'B01001_012E', 'B01001_031E', 'B01001_032E',
            'B01001_033E', 'B01001_034E', 'B01001_035E', 'B01001_036E'
        ],
        # ages 35 - 64
        '35-64': [
            'B01001_013E', 'B01001_014E', 'B01001_015E', 'B01001_016E',
            'B01001_017E', 'B01001_018E', 'B01001_019E', 'B01001_037E',
            'B01001_038E', 'B01001_039E', 'B01001_040E', 'B01001_041E',
            'B01001_042E', 'B01001_043E'
        ],
        # ages 65+
        '65+': [
            'B01001_020E', 'B01001_021E', 'B01001_022E', 'B01001_023E',
            'B01001_024E', 'B01001_025E', 'B01001_044E', 'B01001_045E',
            'B01001_046E', 'B01001_047E', 'B01001_048E', 'B01001_049E'
        ],
    },
    'num_vehicles': {
        '0': ['B08141_002E'],  # No vehicle available
        '1': ['B08141_003E'],  # 1 vehicle available
        '2': ['B08141_004E'],  # 2 vehicles available
        '3+': ['B08141_005E'],  # 3 or more vehicles available
    }
}


CONTROL_NAMES = tuple('_'.join([cat, i]) for cat in CONTROLS.keys()
                      for i in CONTROLS[cat].keys())


class CensusFetchException(Exception):
    pass


class Marginals(object):

    def __init__(self, data):
        self.data = data

    @staticmethod
    def _fetch_from_census(census_key, field_key_list, tract_key, state_key, county_key):
        controls_str = ','.join(field_key_list)
        query = ('http://api.census.gov/data/2015/acs5?get={}'
                 '&for=tract:{}&in=state:{}+county:{}&key={}'
                 ).format(controls_str, tract_key, state_key,
                          county_key, census_key)
        try:
            encoded_response = requests.get(query)
            response = encoded_response.json()
        except Exception:
            print('failed to load marginals for query:\n{}\n'.format(query))
            print('response:\n{}'.format(encoded_response.text))
            raise CensusFetchException()
        control_keys = response[0]
        control_counts = response[1]
        full_controls = dict(zip(control_keys, control_counts))
        return full_controls

    @staticmethod
    def from_census_data(puma_tract_mappings, census_key, state=None, pumas=None):
        """Fetch marginal sums from the census API.

        Args:
            puma_tract_mappings (dict): mappings of the form {
                STATEFP -> state id,
                COUNTYFP -> county id
                PUMA5CE -> puma id
                TRACTCE -> tract id
            }
                of PUMAs to fetch data for

            census_key (unicode): census API key
            state (unicode): state fips code (2-digit id):
                https://www.census.gov/geo/reference/ansi_statetables.html
            pumas (iterable of unicode): a list of pumas to fetch for.  If the
                parameter is not passed in will fetch for all pumas in
                puma_tract_mappings

        Returns:
            Marginals: marginals fetched from the census API

        """
        data = []
        if not pumas or not state:
            raise ValueError('''Please supply a state fips code and a puma.
                    https://www.census.gov/geo/reference/ansi_statetables.html''')

        for line in puma_tract_mappings:
            if state is None or pumas is None\
                    or line['STATEFP'] != state\
                    or (
                            line['PUMA5CE'] not in pumas
                            and
                            line['PUMA5CE'] != pumas
                       ):
                continue
            state_key = line['STATEFP']
            tract_key = line['TRACTCE']
            puma_key = line['PUMA5CE']
            county_key = line['COUNTYFP']
            logging.info('Fetching tract %s', tract_key)
            controls_dict = {}
            success = True
            for cat, control_cat in CONTROLS.items():
                key_list = [key for sublist in list(
                    control_cat.values()) for key in sublist]
                try:
                    full_controls = Marginals._fetch_from_census(
                        census_key, key_list, tract_key, state_key, county_key
                    )
                except CensusFetchException:
                    print('Skip puma {} tract {}'.format(puma_key, tract_key))
                    success = False
                    break
                selected_controls = {
                    key: full_controls[key] for key in full_controls.keys()
                    if key not in ['state', 'public use microdata area']
                }

                for sum_key, sum_cat in control_cat.items():
                    counts = sum([int(selected_controls[key]) for key in sum_cat])
                    controls_dict[cat+'_'+sum_key] = counts
            if success:
                output = [state_key, county_key, puma_key, tract_key]
                for control_name in CONTROL_NAMES:
                    output.append(str(controls_dict[control_name]))
                data.append(output)

        columns = ['STATEFP', 'COUNTYFP', 'PUMA5CE', 'TRACTCE'] + list(CONTROL_NAMES)
        return Marginals(pandas.DataFrame(data, columns=columns))

    @staticmethod
    def from_csv(infile, state=None, puma=None):
        """Load marginals from file.

        Args:
            infile (unicode): path to csv
            state (unicode): state fips code (2-digit)
            puma (unicode): puma code (5-digit)

        Returns:
            Marginals: marginals fetched from a csv file

        """
        data = pandas.read_csv(infile)
        if state is not None and puma is not None:
            data = data[data['STATEFP'] == int(state) and data['PUMA5CE'] == int(puma)]
        return Marginals(data)

    def write(self, outfile):
        """Write marginals to the given file

        Args:
            outfile (unicode): path to write to

        """
        self.data.to_csv(outfile)



# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

"""Utility functions for discretizing values.
"""
from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import sys

import pandas

from doppelganger import inputs


class Preprocessor(object):

    def __init__(
        self, input_to_preprocessor=None, input_to_possible_values=None
    ):
        self.input_to_preprocessor = input_to_preprocessor or {}
        self.input_to_possible_values = input_to_possible_values or {}

    def process_dataframe(self, dataframe, fields, name_map):
        """Extract the data types from the given DataFrames and process them

        Args:
            dataframe (pandas.DataFrame): pums-style data
            fields (iterable): the standard names of fields to extract
            name_map (dict): map of standard name to name within this datasource

        Returns: pandas.DataFrame of the given fields, processed

        """
        cleaned_dataframe = pandas.DataFrame()
        for field_name in fields:
            if field_name in self.input_to_preprocessor:
                procesessor = self.input_to_preprocessor[field_name]
            elif field_name in inputs.NAME_TO_DATATYPE:
                procesessor = inputs.NAME_TO_DATATYPE[field_name].process
            else:
                print('Preprocessor: Unknown data field {}'.format(
                    field_name), file=sys.stderr)
                sys.exit()

            data_type = inputs.NAME_TO_DATATYPE[field_name]
            dirty_name = name_map[field_name]
            if dirty_name in dataframe:
                cleaned_dataframe[data_type.name] = dataframe[
                    dirty_name].apply(procesessor)
            elif dirty_name.upper() in dataframe:
                cleaned_dataframe[data_type.name] = dataframe[
                    dirty_name.upper()].apply(procesessor)
            else:
                print('Missing data field {}'.format(
                    field_name), file=sys.stderr)
        return cleaned_dataframe

    def get_possible_values(self, field):
        field = inputs.NAME_TO_DATATYPE[field]
        if field.name in self.input_to_possible_values:
            return self.input_to_possible_values[field.name]
        return field.possible_values

    @staticmethod
    def from_config(config):
        """Load a preprocessor from a config.

        If unspecified, assumes that data fields use PUMS names.
        """
        field_to_preprocessor = {}
        field_to_possible_values = {}
        for field in config:
            if 'bins' in config[field]:
                labels, preprocessor = inputs.generate_binning_preprocessor(
                    config[field]['bins']
                )
                field_to_preprocessor[field] = preprocessor
                field_to_possible_values[field] = labels
        return Preprocessor(field_to_preprocessor, field_to_possible_values)



# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from collections import defaultdict, namedtuple
import numpy as np
import pandas

from doppelganger.listbalancer import (
    balance_multi_cvx, discretize_multi_weights
)
from doppelganger import inputs


HIGH_PASS_THRESHOLD = .1  # Filter controls which are present in less than 10% of HHs

# These are the minimum fields needed to allocate households
DEFAULT_PERSON_FIELDS = {
    inputs.STATE,
    inputs.PUMA,
    inputs.SERIAL_NUMBER,
    inputs.AGE,
    inputs.SEX,
    inputs.PERSON_WEIGHT,
}

DEFAULT_HOUSEHOLD_FIELDS = {
    inputs.STATE,
    inputs.PUMA,
    inputs.SERIAL_NUMBER,
    inputs.NUM_PEOPLE,
    inputs.HOUSEHOLD_WEIGHT,
}


CountInformation = namedtuple('CountInformation', ['tract', 'count'])


class HouseholdAllocator(object):

    @staticmethod
    def from_csvs(households_csv, persons_csv):
        """Load saved household and person allocations.

        Args:
            households_csv (unicode): path to households file
            persons_csv (unicode): path to persons file

        Returns:
            HouseholdAllocator: allocated persons & households_csv

        """
        allocated_households = pandas.read_csv(households_csv)
        allocated_persons = pandas.read_csv(persons_csv)
        return HouseholdAllocator(allocated_households, allocated_persons)

    @staticmethod
    def from_cleaned_data(marginals, households_data, persons_data):
        """Allocate households based on the given data.

        marginals (Marginals): controls to match when allocating
        households_data (CleanedData): data about households.  Must contain
            DEFAULT_HOUSEHOLD_FIELDS.
        persons_data (CleanedData): data about persons.  Must contain
            DEFAULT_PERSON_FIELDS.
        """
        for field in DEFAULT_HOUSEHOLD_FIELDS:
            assert field.name in households_data.data, \
                'Missing required field {}'.format(field.name)
        for field in DEFAULT_PERSON_FIELDS:
            assert field.name in persons_data.data, \
                'Missing required field {}'.format(field.name)

        households, persons = HouseholdAllocator._format_data(
            households_data.data, persons_data.data)
        allocated_households, allocated_persons = \
            HouseholdAllocator._allocate_households(households, persons, marginals)
        return HouseholdAllocator(allocated_households, allocated_persons)

    def __init__(self, allocated_households, allocated_persons):

        self.allocated_households = allocated_households
        self.allocated_persons = allocated_persons
        self.serialno_to_counts = defaultdict(list)
        for _, row in self.allocated_households.iterrows():
            serialno = row[inputs.SERIAL_NUMBER.name]
            tract = row[inputs.TRACT.name]
            count = int(row[inputs.COUNT.name])
            self.serialno_to_counts[serialno].append(CountInformation(tract, count))

    def get_counts(self, serialno):
        """Return the information about weights for a given serial number.

        A household is repeated for a certain number of times for each tract.
        This returns a list of (tract, repeat count).  The repeat count
        indicates the number of times this serial number should be repeated in
        this tract.

        Args:
            seriano (unicode): the household's serial number

        Returns:
            list(CountInformation): the weighted repetitions for this serialno
        """
        return self.serialno_to_counts[serialno]

    def write(self, household_file, person_file):
        """Write allocated households and persons to the given files

        Args:
            household_file (unicode): path to write households to
            person_file (unicode): path to write persons to

        """
        self.allocated_households.to_csv(household_file)
        self.allocated_persons.to_csv(person_file)

    @staticmethod
    def _filter_sparse_columns(df, cols):
        ''' Filter out variables who are are so sparse they would break the solver.
        Columns are assumed to be of an indicator type (0/1)
        Args
            df (pandas.DataFrame): dataframe to filter
            cols (list(str)): column names

        Returns
            filtered column list (list(str))
        '''
        return df[cols]\
            .loc[:, df[cols].sum()/float(len(df)) > HIGH_PASS_THRESHOLD]\
            .columns.tolist()

    @staticmethod
    def _allocate_households(households, persons, tract_controls):

        # Only take nonzero weights
        households = households[households[inputs.HOUSEHOLD_WEIGHT.name] > 0]

        # Initial weights from PUMS
        w = households[inputs.HOUSEHOLD_WEIGHT.name].as_matrix().T

        allocation_inputs = [inputs.NUM_PEOPLE, inputs.NUM_VEHICLES]  # Hard-coded for now
        # Prepend column name to bin name to prevent bin collision
        hh_columns = []
        for a_input in allocation_inputs:
            subset_values = households[a_input.name].unique().tolist()
            hh_columns += HouseholdAllocator._str_broadcast(a_input.name, subset_values)

        hh_columns = HouseholdAllocator._filter_sparse_columns(households, hh_columns)

        hh_table = households[hh_columns].as_matrix()

        A = tract_controls.data[hh_columns].as_matrix()
        n_tracts, n_controls = A.shape
        n_samples = len(households.index.values)

        # Control importance weights
        # < 1 means not important (thus relaxing the constraint in the solver)
        mu = np.mat([1] * n_controls)

        w_extend = np.tile(w, (n_tracts, 1))
        mu_extend = np.mat(np.tile(mu, (n_tracts, 1)))
        B = np.mat(np.dot(np.ones((1, n_tracts)), A)[0])

        # Our trade-off coefficient gamma
        # Low values (~1) mean we trust our initial weights, high values
        # (~10000) mean want to fit the marginals.
        gamma = 100.

        # Meta-balancing coefficient
        meta_gamma = 100.

        hh_weights = balance_multi_cvx(
            hh_table, A, B, w_extend, gamma * mu_extend.T, meta_gamma
        )

        # We're running discretization independently for each tract
        tract_ids = tract_controls.data['TRACTCE'].values
        total_weights = np.zeros(hh_weights.shape)
        sample_weights_int = hh_weights.astype(int)
        discretized_hh_weights = discretize_multi_weights(hh_table, hh_weights)
        total_weights = sample_weights_int + discretized_hh_weights

        # Extend households and add the weights and ids
        households_extend = pandas.concat([households] * n_tracts)
        households_extend[inputs.COUNT.name] = total_weights.flatten().T
        tracts = np.repeat(tract_ids, n_samples)
        households_extend[inputs.TRACT.name] = tracts

        return households_extend, persons

    @staticmethod
    def _str_broadcast(string, list1):
        return ['_'.join([string, element]) for element in list1]

    @staticmethod
    def _format_data(households_data, persons_data):
        hh_size = pandas.get_dummies(households_data[inputs.NUM_PEOPLE.name])
        # Prepend column name to bin name to prevent bin collision
        hh_size.columns = HouseholdAllocator\
            ._str_broadcast(inputs.NUM_PEOPLE.name, hh_size.columns.tolist())

        hh_vehicles = pandas.get_dummies(households_data[inputs.NUM_VEHICLES.name])
        hh_vehicles.columns = HouseholdAllocator\
            ._str_broadcast(inputs.NUM_VEHICLES.name, hh_vehicles.columns.tolist())
        households_data = pandas.concat([households_data, hh_size, hh_vehicles], axis=1)

        hp_ages = pandas.get_dummies(persons_data[inputs.AGE.name])
        hp_ages.columns = HouseholdAllocator\
            ._str_broadcast(inputs.AGE.name, list(inputs.AGE.possible_values))
        persons_data = pandas.concat([persons_data, hp_ages], axis=1)

        persons_trimmed = persons_data[[
                inputs.SERIAL_NUMBER.name
            ] + hp_ages.columns.tolist()
            ]

        # Get counts we need
        persons_trimmed = persons_trimmed.groupby(inputs.SERIAL_NUMBER.name).sum()

        households_trimmed = households_data[[
            inputs.SERIAL_NUMBER.name,
            inputs.NUM_PEOPLE.name,
            inputs.NUM_VEHICLES.name,
            inputs.HOUSEHOLD_WEIGHT.name
        ] + hh_size.columns.tolist()
          + hh_vehicles.columns.tolist()
        ]

        # Merge
        households_out = pandas.merge(
            households_trimmed, persons_trimmed, how='inner',
            left_on=inputs.SERIAL_NUMBER.name, right_index=True, sort=True
        )

        persons_out = persons_data[[
            inputs.SERIAL_NUMBER.name,
            inputs.SEX.name,
            inputs.AGE.name
        ]]

        return households_out, persons_out



# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

"""Library for modeling generative typed bayesian networks.

Each model consists of multiple networks, one for each type of input data.
For instance, a person model might have a different network for persons of
each household type. This allows the model to learn transition probabilities
for data of each type.

`SegmentedData` helps train this typed network structure by allowing users to
specify a segmentation function to segment the training data by type.
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
from builtins import range, str

from collections import defaultdict, Counter
import json
import itertools
import sys

import pandas
from pomegranate import BayesianNetwork


def default_segmenter(x):
    return 'one_segment'


class SegmentedData(object):
    """Segmented data for use with the segemented BayesianNetworkModel.

    Like the model itself, training data uses a mapping of type -> data.

    """

    def __init__(self, type_to_data, segmenter=None):
        self.type_to_data = type_to_data
        self.segmenter = segmenter

    @staticmethod
    def from_data(cleaned_data, fields, weight_field=None, segmenter=None):
        """Input more data.

        Args:
            cleaned_data (CleanedData): data to train on
            segmenter: function mapping a dict of data to a type for
                segmentation
            weight_field (unicode): Name of the int field that shows how much
                this row  of data should be weighted.
        """
        segmenter = segmenter or default_segmenter
        type_to_data = defaultdict(list)
        for _, row in cleaned_data.data.iterrows():
            type_ = segmenter(row)
            weight = row[weight_field] if weight_field else 1
            cleaned_row = tuple(row[fields])
            for _ in range(weight):
                type_to_data[type_].append(cleaned_row)
        return SegmentedData(type_to_data, segmenter)

    def num_rows_data(self):
        return sum(len(data) for data in self.type_to_data.values())

    def types(self):
        return self.type_to_data.keys()


class BayesianNetworkModel(object):
    """A typed Bayesian network model.

    This bayesian network model as a fixed list of nodes passed in at creation.
    It holds a separate network for each user-defined type.
    """

    def __init__(self, type_to_network, fields, segmenter=None):
        self.type_to_network = type_to_network
        self.fields = fields
        self.distribution_cache = {}
        self.segmenter = segmenter or default_segmenter

    @staticmethod
    def from_file(filename, segmenter=None):
        with open(filename) as infile:
            json_string = infile.read()
            return BayesianNetworkModel.from_json(json_string, segmenter)

    def write(self, outfilename):
        with open(outfilename, 'w') as outfile:
            json_string = self.to_json()
            outfile.write(json_string)

    def to_json(self):
        blob = {'fieldnames': self.fields}
        blob['type_to_network'] = {
            type_: json.loads(network.to_json()) for type_, network in self.type_to_network.items()
        }
        return json.dumps(blob, indent=4, sort_keys=True)

    @staticmethod
    def _df_from_conditional(probabilities):
        """
        Helper method to extract a probability table from pomegranate's json
        format for conditional probability distributions.
        """
        state_map = defaultdict(dict)
        for row in probabilities:
            evidence = tuple(row[:-2])
            value = row[-2]
            probability = float(row[-1])
            state_map[evidence][value] = probability
        return pandas.DataFrame.from_dict(state_map).transpose()

    @staticmethod
    def _df_from_discrete(probabilities):
        """
        Helper method to extract a probability table from pomegranate's json
        format for discrete distributions.
        """
        return pandas.DataFrame(probabilities)

    def probabilities_as_dataframes(self):
        """Create dataframes for each node in each bayesian network.

        Returns:
            dict {str -> list(DataFrame)} Dictionary from segment name to
                a list of DataFrames, one for each state of the distribution
                for that segment.  For nodes with no ancestors, the dataframe
                is just columns with the probability of each value.  For nodes
                with ancestors, the dataframe frame's rows labels are evidence
                and column labels are the values, with the cell representing
                the probability of the value given the evidence.
        """
        segment_to_states = {}
        for segment, network in self.type_to_network.items():
            state_to_dataframes = []
            for state in network.states:
                # The one to access the transition probabilities directly is
                # via the json output
                distribution = json.loads(str(state))['distribution']
                if distribution['name'] == 'ConditionalProbabilityTable':
                    probabilities = BayesianNetworkModel._df_from_conditional(
                        distribution['table'])
                else:
                    probabilities = BayesianNetworkModel._df_from_discrete(
                        distribution['parameters']
                    )
                state_to_dataframes.append(probabilities)
            segment_to_states[segment] = state_to_dataframes
        return segment_to_states

    @staticmethod
    def from_json(json_string, segmenter=None):
        """Create BayesianNetworkModel from the given json blob in string format

        Args:
            json_string (unicode): the string created by `from_json`

        Returns:
            BayesianNetworkModel: generative model equivalent to stored model
        """
        json_blob = json.loads(json_string)
        type_to_network = {}
        for type_, network_json in json_blob['type_to_network'].items():
            type_to_network[type_] = BayesianNetwork.from_json(json.dumps(network_json))
        fields = list(json_blob['fieldnames'])
        return BayesianNetworkModel(type_to_network, fields, segmenter)

    @staticmethod
    def train(input_data, structure, fields, prior_data=None):
        """Creates bayesian networks from the given data with the given structure.

        The given data cannot contain any missing data. If called multiple
        times, the old model will be replaced.  To update the model with new
        data, see `update`.

        Args:
            input_data (SegmentedData): typed data to train on
            structure (iterable(iterable)): structure as returned from
                    define_bayes_net_structure
            fields (list(unicode)): field names to learn
            prior_data (list(data)): optional list of training samples to use
                    as a prior for each network.

        Return:
            BayesianNetworkModel: A predictive model training on the given data

        """
        type_to_network = {}
        for type_, data in input_data.type_to_data.items():
            if prior_data is not None:
                # Make defensive copy
                data = list(data) + list(prior_data)
            bayesian_network = BayesianNetwork.from_structure(data, structure)
            type_to_network[type_] = bayesian_network
        return BayesianNetworkModel(type_to_network, fields, segmenter=input_data.segmenter)

    def log_likelihood(self, training_data):
        """Compute the log likelihood of the given data given the model

        Compute the log likehood of the data for each type based on the
            Bayesian network for that type.  Assumes all data rows are
            independent.

        Args:
            input_data (SegmentedData): data whose likelihood to compute

        Returns:
            {type -> log likelihood}: The likelihood of the data for each
                type of data
        """
        type_to_likelihood = {}
        for type_, data in training_data.type_to_data.items():
            network = self.type_to_network[type_]
            data_counter = Counter(tuple(x) for x in data)
            log_likelihood = 0.0
            for data_row, count in data_counter.items():
                try:
                    log_likelihood += count * network.log_probability(data_row)
                except KeyError:
                    message = 'Data with zero likelihood {}'.format(data_row)
                    print(message, file=sys.stderr)
                    log_likelihood = float('-inf')
                    break
            type_to_likelihood[type_] = log_likelihood
        return type_to_likelihood

    def update(self, input_data, max_iterations=1, inertia=0.0):
        """Updates the distribution of a trained network based on new data.

        Missing values are accepted.  Missing values are filled in using MLE
        from the old distribution, and then the distribution is mutated to
        include the new data.

        Args:
            input_data (SegmentedData): typed data to train on
            max_iterations: (int): max number of iteratations of
                Expectation-Maximization to fill in missing data. Will run
                until hitting max_iterations or convergence, whichever happens
                first.
            inertia (float): The weight of the previous parameters of the
                model. The new parameters will roughly be old_param*inertia +
                new_param*(1-inertia), so an inertia of 0 means ignore the old
                parameters, whereas an inertia of 1 means ignore the new
                parameters. Default is 0.0.

        Return:
            BayesianNetworkModel: self, a predictive model training on the
                given data

        """
        def data_equals(old, new):
            if old is None or new is None:
                return False
            assert len(old) == len(new)
            for i in range(len(old)):
                # Compare as tuple because numpy arrays return an array
                # of bools instead of a bool on comparison.
                if tuple(old[i]) != tuple(new[i]):
                    return False
            return True

        # For each data-type, use EM to learn missing fields and update the
        # model
        for type_, data in input_data.type_to_data.items():
            data_previous = None
            data_new = None
            iteration = 0
            while not data_equals(data_previous, data_new):
                data_previous = data_new
                if iteration >= max_iterations:
                    break
                iteration += 1
                # Make a copy of the original data with the missing fields
                data_new = [list(row) for row in data]
                bayesian_network = self.type_to_network[type_]
                # Fill in missing fields
                data_new = bayesian_network.predict(data_new)
                # Update the model
                bayesian_network.fit(data_new, inertia=inertia)
        return self

    def generate(self, type_, evidence, count=1):
        """Sample from the network based on the given evidence

        Args:
            type_: user-defined type that will determine the network to use
            evidence ((field name, value), ...): any prior observed data. The
                    field names must be in the fields supplied on model
                    creation.
            count (int): the number of samples to generate
        Returns:
            tuple of data sampled, one element for each of the fields
                    supplied on model creation.

        """
        if (type_, evidence) in self.distribution_cache:
            distributions = self.distribution_cache[(type_, evidence)]
        else:
            try:
                evidence_translated = {
                    str(self.fields.index(field)): value
                    for field, value in evidence
                }
            except ValueError:
                raise ValueError('Evidence supplied not in model fields')
                # When pomegranate supports sampling directly from the BN we
                # will use that. See github issue
                # https://github.com/jmschrei/pomegranate/issues/231
            distributions = self.type_to_network[
                type_].predict_proba(evidence_translated)
            self.distribution_cache[(type_, evidence)] = distributions

        generated = tuple(
            tuple(distribution.sample() for distribution in distributions) for _ in range(count)
        )
        return generated


def define_bayes_net_structure(nodes, edges):
    """Create a bayes network based on the given configuration

    Args:
        nodes iterable(unicode): names of nodes
        edges dict(unicode -> unicode): edges in the form of parent -> child

    Returns: bayes net structure currently as a list where element i is a list
        of the indices of parents of i
    """
    child_to_parents = defaultdict(set)
    for parent, children in edges.items():
        for child in children:
            child_to_parents[child].add(parent)
    node_to_index = {name: i for i, name in enumerate(nodes)}
    structure = []
    for child in nodes:
        structure.append(tuple(node_to_index[parent] for parent in child_to_parents[child]))
    return tuple(tuple(s) for s in structure)


def generate_laplace_prior_data(fields, preprocessor):
    """Create training data for Laplace smoothing

    Generate data needed to apply Laplace smoothing to Bayesian network
    training. The value returned here can be applied to
    `BayesianNetworkModel.train`'s `prior_data` parameter to give each possible
    combination of field values an equal prior.

    Args:
        fields (iterable(string)): the names of all fields in the training data
        preprocessor (Preprocessor): preprocessor used for processing the
            training data, needed because this determines the data's possible
            values.

    Returns: (iterable(value list)) a list of data for use by
        `BayesianNetworkModel.train`'s `prior_data` parameter

    """
    all_values = (preprocessor.get_possible_values(field) for field in fields)
    return set(itertools.product(*all_values))



# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

from .allocation import HouseholdAllocator
from .bayesnets import SegmentedData, BayesianNetworkModel
from .config import Configuration
from .datasource import PumsData, CleanedData, DirtyDataSource
from .marginals import Marginals
from .preprocessing import Preprocessor
from .populationgen import Population

# Enumerate exports, to make the linter happy.
__all__ = [
    HouseholdAllocator, SegmentedData, BayesianNetworkModel, Configuration,
    PumsData, CleanedData, Marginals, Population, Preprocessor, DirtyDataSource,
]



# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import unittest
import pandas

from doppelganger import datasource, inputs, Preprocessor


class DataSourceTest(unittest.TestCase):

    def _mock_dirty_household_input(self):
        return [
            {'serialno': 'a', 'np': '1', 'veh': '1', 'fincp': ''},
            {'serialno': 'b', 'np': '2', 'veh': '6', 'fincp': '0070000'},
            {'serialno': 'c', 'np': '1', 'veh': '', 'fincp': '0000000'}
        ]

    def _mock_dirty_people_input(self, weight=1):
        def mock_person(serialno, age, sex, income):
            return {
                'serialno': serialno,
                'agep': age,
                'sex': sex,
                'wagp': income,
                'pwgtp': weight
            }
        return [
            mock_person('a', '15', '1', ''),
            mock_person('b', '50', '2', '0070000'),
            mock_person('b', '70', '1', '0015000'),
            mock_person('c', '25', '1', '0000000')
        ]

    def test_clean_data(self):
        pums_data = datasource.PumsData(
            pandas.DataFrame(self._mock_dirty_household_input())
        )
        cleaned = pums_data.clean([
            inputs.SERIAL_NUMBER.name,
            inputs.NUM_PEOPLE.name,
            inputs.NUM_VEHICLES.name,
            inputs.HOUSEHOLD_INCOME.name
        ],
            Preprocessor())
        actual = cleaned.data.loc[1].to_dict()
        expected = {
            inputs.SERIAL_NUMBER.name: 'b',
            inputs.NUM_PEOPLE.name: '2',
            inputs.NUM_VEHICLES.name: '3+',
            inputs.HOUSEHOLD_INCOME.name: '40000+'
        }
        self.assertDictEqual(actual, expected)

    def test_clean_data_one_field(self):
        pums_data = datasource.PumsData(
            pandas.DataFrame(self._mock_dirty_household_input())
        )
        cleaned = pums_data.clean([inputs.NUM_PEOPLE.name], Preprocessor())
        actual = cleaned.data.loc[1].to_dict()
        expected = {
            inputs.NUM_PEOPLE.name: '2',
        }
        self.assertDictEqual(actual, expected)



# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import unittest
import numpy as np

from doppelganger import listbalancer


class ListBalancerTests(unittest.TestCase):

    def _mock_list_consistent(self):
        hh_table = np.mat([
            [1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0],
            [0, 0, 1, 0, 1],
            [0, 0, 0, 1, 1]
        ])
        A = np.mat([
            81.,
            101.,
            151.,
            429.,
            580.,
        ])
        w = np.matrix([[
            81.,
            101.,
            151.,
            429.
        ]])
        _, n_controls = hh_table.shape

        expected_weights = np.matrix([
            [81.],
            [101.],
            [151.],
            [429.]
        ])

        mu = np.mat([1] * n_controls)

        return (hh_table, A, w, mu, expected_weights)

    def _mock_list_relaxed(self):
        hh_table = np.mat([
            [1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0],
            [0, 0, 1, 0, 1],
            [0, 0, 0, 1, 1]
        ])
        A = np.mat([
            81.,
            101.,
            151.,
            429.,
            580.,
        ])
        w = np.matrix([[
            81.,
            101.,
            151.,
            429.
        ]])
        _, n_controls = hh_table.shape

        expected_weights = np.matrix([
            [45.],
            [52.],
            [65.],
            [98.]
        ])

        mu = np.mat([1] * n_controls)

        return (hh_table, A, w, mu, expected_weights)

    def _mock_list_inconsistent(self):
        hh_table = np.mat([
            [1, 0, 0, 1, 0],
            [0, 1, 0, 1, 1],
            [0, 0, 1, 2, 1],
            [0, 0, 1, 1, 2]
        ])
        A = np.mat([
            81.,
            101.,
            151.,
            429.,
            299.,
        ])
        w = np.matrix([
            [79.,
             99.,
             101.,
             49.]
        ])
        _, n_controls = hh_table.shape
        mu = np.mat([1] * n_controls)

        expected_weights = np.matrix([
            [70.66,
             88.66,
             85.47,
             45.72]
        ])

        return (hh_table, A, w, mu, expected_weights)

    def _mock_list_infeasible(self):
        hh_table = np.mat([
            [1, 0, 0, 1, 0],
            [0, 1, 0, 1, 1],
            [0, 0, 1, 2, 1],
            [0, 0, 1, 1, 2]
        ])
        A = np.mat([
            81.,
            101.,
            151.,
            429.,
            299.,
        ])
        w = np.matrix([
            [1000.,
             0.,
             0.,
             0.]
        ])
        _, n_controls = hh_table.shape
        mu = np.mat([1] * n_controls)

        expected_weights = np.matrix([
            [100.,    0.,    0.,    0.]
        ])

        return (hh_table, A, w, mu, expected_weights)

    def _mock_list_infeasible_marginal(self):
        hh_table = np.mat([
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1],
            [0, 0, 0, 1],
            [0, 0, 1, 1],
        ])
        A = np.mat([
            [0., 324., 0, 0.],
            [0., 357., 0, 0.],
            [0., 138., 0., 0.],
            [0., 183., 0., 0.],
        ]).T
        w = np.matrix([
            [0., 0., 0., 0., 0.],
            [79., 99., 101., 49., 200],
            [0., 0., 0., 0., 0],
            [0., 0., 0., 0., 0],
        ])
        _, n_controls = hh_table.shape
        mu = np.mat(np.ones((2, n_controls)))

        expected_weights = np.matrix([
            [0., 0., 0., 0., 0.],
            [340.46, 58.09, 69.56, 33.75, 80.83],
            [0., 0., 0., 0., 0.],
            [0., 0., 0., 0., 0.],
        ])

        return (hh_table, A, w, mu, expected_weights)

    def _mock_hh_weights(self):
        hh_table = np.mat([
            [1, 0, 0, 1, 0],
            [0, 1, 0, 1, 1],
            [0, 0, 1, 2, 1],
            [0, 0, 1, 1, 2]
        ])
        hh_weights = np.mat([
            [80.79, 100.6, 100.7, 49.32],
            [80.79, 100.6, 100.7, 49.32],
        ])
        expected_hh_discretized = np.mat([
            [1, 1, 1, 0],
            [1, 1, 1, 0],
        ])
        return hh_table, hh_weights, expected_hh_discretized

    def _mock_hh_weights_zeroed(self):
        hh_table = np.mat([
            [1, 0, 0, 1, 0],
            [0, 1, 0, 1, 1],
            [0, 0, 1, 2, 1],
            [0, 0, 1, 1, 2]
        ])
        hh_weights = np.mat([
            [0., 0., 0., 0.],
            [80.79, 100.6, 100.7, 49.32],
            [0., 0., 0., 0.],
        ])
        expected_hh_discretized = np.mat([
            [0, 0, 0, 0],
            [1, 1, 1, 0],
            [0, 0, 0, 0],
        ])
        return hh_table, hh_weights, expected_hh_discretized

    def test_balance_cvx(self):
        hh_table, A, w, _, expected_weights = self._mock_list_consistent()
        hh_weights = listbalancer.balance_cvx(hh_table, A, w)
        np.testing.assert_allclose(
            hh_weights, expected_weights, rtol=0.01, atol=0)

    def test_balance_cvx_relaxed(self):
        hh_table, A, w, mu, expected_weights = self._mock_list_relaxed()
        hh_weights, _ = listbalancer.balance_cvx(hh_table, A, w, mu)
        np.testing.assert_allclose(
            hh_weights, expected_weights, rtol=0.01, atol=0)

    def test_balance_multi_cvx(self):
        hh_table, A, w, mu, expected_weights = self._mock_list_inconsistent()

        # Extend the data
        n_tracts = 10
        A_extend = np.mat(np.tile(A, (n_tracts, 1)))
        w_extend = np.mat(np.tile(w, (n_tracts, 1)))
        mu_extend = np.mat(np.tile(mu, (n_tracts, 1)))
        B = np.mat(np.dot(np.ones((1, n_tracts)), A_extend)[0])
        expected_weights_extend = np.mat(
            np.tile(expected_weights, (n_tracts, 1)))
        gamma = 1000.
        meta_gamma = 1000.
        hh_weights = listbalancer.balance_multi_cvx(
            hh_table, A_extend, B, w_extend, gamma * mu_extend.T, meta_gamma)
        np.testing.assert_allclose(
            hh_weights, expected_weights_extend, rtol=0.01, atol=0)

    def test_balance_multi_cvx_infeasible(self):
        hh_table, A, w, mu, expected_weights = self._mock_list_infeasible()

        # Extend the data
        n_tracts = 10
        A_extend = np.mat(np.tile(A, (n_tracts, 1)))
        w_extend = np.mat(np.tile(w, (n_tracts, 1)))
        expected_weights_extend = np.mat(np.tile(expected_weights, (n_tracts, 1)))
        mu_extend = np.mat(np.tile(mu, (n_tracts, 1)))
        B = np.mat(np.dot(np.ones((1, n_tracts)), A_extend)[0])
        gamma = 10.
        meta_gamma = 1000.
        hh_weights = listbalancer.balance_multi_cvx(
            hh_table, A_extend, B, w_extend, gamma * mu_extend.T, meta_gamma
        )
        np.testing.assert_allclose(
            hh_weights, expected_weights_extend, rtol=0.01, atol=0)

    def test_balance_multi_trust_initial(self):
        hh_table, A, w, mu, _ = self._mock_list_inconsistent()
        B = np.mat(np.dot(np.ones((1, 1)), A)[0])
        gamma = 1.
        hh_weights = listbalancer.balance_multi_cvx(
            hh_table, A, B, w, gamma * mu.T
        )
        np.testing.assert_allclose(
            hh_weights, w, rtol=0.05, atol=0)

    def test_balance_multi_trust_controls(self):
        hh_table, A, w, mu, expected_weights = self._mock_list_consistent()
        B = np.mat(np.dot(np.ones((1, 1)), A)[0])
        gamma = 100000.
        hh_weights = listbalancer.balance_multi_cvx(
            hh_table, A, B, w, gamma * mu.T
        )
        np.testing.assert_allclose(
            hh_weights, expected_weights.T, rtol=0.05, atol=0)

    def test_balance_multi_zero_marginal(self):
        hh_table, A, w, mu, expected_weights = \
            self._mock_list_infeasible_marginal()
        n_tracts = A.shape[0]
        B = np.mat(np.dot(np.ones((1, n_tracts)), A)[0])
        gamma = 10000.
        hh_weights = listbalancer.balance_multi_cvx(
            hh_table, A, B, w, gamma * mu.T
        )
        np.testing.assert_allclose(
            hh_weights, expected_weights, rtol=0.05, atol=0)

    def test_discretize_multi_weights(self):
        hh_table, hh_weights, expected_hh_discretized = self._mock_hh_weights()
        hh_discretized = listbalancer.discretize_multi_weights(
            hh_table, hh_weights)
        np.testing.assert_array_equal(
            hh_discretized, expected_hh_discretized)

    def test_discretize_multi_zero_weights(self):
        hh_table, hh_weights, expected_hh_discretized = self._mock_hh_weights_zeroed()
        hh_discretized = listbalancer.discretize_multi_weights(
            hh_table, hh_weights)
        np.testing.assert_array_equal(
            hh_discretized, expected_hh_discretized)



#!/usr/bin/env python

# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

from setuptools import setup

setup(
    name="doppelganger",
    version="0.1.2",
    description='Population synthesis library',
    author='Kat Busch, Kael Greco and contributors',
    author_email='doppelganger@sidewalklabs.com',
    packages=['doppelganger'],
    install_requires=[
        'cvxpy==0.4.8',
        'numpy>=1.11.0',
        'pandas>=0.19.0',
        'pomegranate==0.7.1',
        'requests>=2.0.0',
        'six>=1.10.0',
        'future>=0.16.0',
        'psycopg2==2.7.3.1',
        'enum34',
    ],
    extras_require={
        'tests': [
            'flake8>=3.5.0',
            'mock>=2.0.0',
            'nose>=1.3.4',
            'coveralls>=1.1',
            'pytest',
        ],
    },

)



# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import pandas

from doppelganger import inputs


class Population(object):

    @staticmethod
    def from_csvs(persons_infile, households_infile):
        """Load generated population from file.

        Args:
            persons_infile (unicode): path to persons csv
            households_infile (unicode): path to households csv

        Returns:
            Population: generated population

        """
        generated_people = pandas.read_csv(persons_infile)
        generated_households = pandas.read_csv(households_infile)
        return Population(generated_people, generated_households)

    def __init__(self, generated_people, generated_households):
        self.generated_people = generated_people
        self.generated_households = generated_households

    @staticmethod
    def _extract_person_evidence(allocated_rows, fields, segmenter, household_allocator):
        """Creates a (python) generator for bayesian network evidence for persons that yields
         (serial number, evidence, segment, tract, count)
        """
        for _, row in allocated_rows.iterrows():
            serialno = row[inputs.SERIAL_NUMBER.name]
            evidence = tuple((field, row[field]) for field in fields)
            segment = segmenter(row)
            # Draw repeat information for persons from from the allocator
            count_info = household_allocator.get_counts(serialno)
            for tract, count in count_info:
                yield serialno, evidence, segment, tract, count

    @staticmethod
    def _extract_household_evidence(allocated_rows, fields, segmenter, _):
        """Creates a (python) generator for bayesian network evidence for households that yields
         (serial number, evidence, segment, tract, count)
        """
        for _, row in allocated_rows.iterrows():
            serialno = row[inputs.SERIAL_NUMBER.name]
            evidence = tuple((field, row[field]) for field in fields)
            segment = segmenter(row)
            # Households store their repeat information directly
            yield serialno, evidence, segment, row[inputs.TRACT.name], row['count']

    @staticmethod
    def _generate_from_model(household_allocator, data, model, fields, evidence_fn):
        """Generate the given fields of the given data generated by the
        given model
        """
        results = []
        for serialno, evidence, segment, tract, count in evidence_fn(
                data,
                fields,
                model.segmenter,
                household_allocator
        ):
            generated_rows = model.generate(segment, evidence, count=count)
            for repeat_id, row in enumerate(generated_rows):
                household_id = '{}-{}-{}'.format(tract, serialno, repeat_id)
                results.append((household_id, tract, serialno, repeat_id) + row)

        column_names = [inputs.HOUSEHOLD_ID.name, inputs.TRACT.name, inputs.SERIAL_NUMBER.name,
                        inputs.REPEAT_INDEX.name] + list(model.fields)
        results_dataframe = pandas.DataFrame(results, columns=column_names)
        return results_dataframe

    @staticmethod
    def generate(household_allocator, person_model, household_model):
        """Create all the persons and households for this population

        Args:
            household_allocator (HouseholdAllocator): allocated households
            person_model (BayesianNetworkNodel): optional generative model
            household_model (BayesianNetworkNodel): optional generative model

        Returns: Population from the given model
        """

        persons = Population._generate_from_model(
            household_allocator, household_allocator.allocated_persons,
            person_model, [inputs.AGE.name, inputs.SEX.name], Population._extract_person_evidence
        )
        households = Population._generate_from_model(
            household_allocator, household_allocator.allocated_households,
            household_model, [inputs.NUM_PEOPLE.name], Population._extract_household_evidence
        )
        return Population(persons, households)

    def write(self, persons_outfile, households_outfile):
        """Write population to the given file

        Args:
            outfile (unicode): path to write to

        """
        self.generated_people.to_csv(persons_outfile)
        self.generated_households.to_csv(households_outfile)



from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import unittest
import os
import mock
from doppelganger.scripts import download_allocate_generate
from doppelganger import config, CleanedData, inputs
import pandas


class TestDownloadAllocateGenerate(unittest.TestCase):

    def _mock_config(self):
        conf = {
            "person_fields": [
                "age",
                "sex",
                "individual_income"
            ],
            "household_fields": [
                "num_people",
                "household_income",
                "num_vehicles"
            ],
            "preprocessing": {
                "individual_income": {
                    "bins": [
                        0,
                        20000,
                        40000,
                        80000,
                        100000
                    ]
                }
            },
            "network_config_files": {
                "person": os.path.dirname(__file__)
                + "/../examples/sample_data/sample_person_bn.json",
                "household": os.path.dirname(__file__)
                + "/../examples/sample_data/sample_household_bn.json"
            },
            "version": "0"
        }
        return config.Configuration.from_json(conf)

    _mock_params = {
                    'state_id': '01',
                    'puma_id': '00001',
                    'output_dir': '.',
                    'db_host': 'host1',
                    'db_database': 'database1',
                    'db_schema': 'schema1',
                    'db_table': 'table1',
                    'db_user': 'user1',
                    'db_password': 'password1',
                    'census_api_key': 'census_key1',
                    'puma_tract_mappings': 'puma_tract_mappings',
                }

    @mock.patch('pandas.DataFrame.to_csv')
    @mock.patch('doppelganger.scripts.download_allocate_generate.CleanedData')
    @mock.patch('doppelganger.scripts.download_allocate_generate.os.path.exists')
    @mock.patch('doppelganger.scripts.download_allocate_generate.fetch_pums_data')
    def test_download_and_load_pums_data_download(self, mock_fetch_pums_data, mock_exists,
                                                  mock_CleanedData, mock_pandas_to_csv):
        '''Verify fetch_pums_data is called with the proper arguments if local pums files aren't
        found
        '''
        mock_fetch_pums_data.return_value = (CleanedData(pandas.DataFrame()),
                                             CleanedData(pandas.DataFrame()))
        configuration = self._mock_config()
        mock_exists.return_value = False

        download_allocate_generate.download_and_load_pums_data(
            output_dir=self._mock_params['output_dir'],
            state_id=self._mock_params['state_id'],
            puma_id=self._mock_params['puma_id'],
            configuration=configuration,
            db_host=self._mock_params['db_host'],
            db_database=self._mock_params['db_database'],
            db_schema=self._mock_params['db_schema'],
            db_user=self._mock_params['db_user'],
            db_password=self._mock_params['db_password']
        )
        mock_fetch_pums_data.assert_called()
        mock_fetch_pums_data.assert_called_with(
                        state_id='01',
                        puma_id='00001',
                        configuration=configuration,
                        db_host='host1',
                        db_database='database1',
                        db_schema='schema1',
                        db_user='user1',
                        db_password='password1'
                    )

    @mock.patch('pandas.DataFrame.to_csv')
    @mock.patch('doppelganger.scripts.download_allocate_generate.CleanedData')
    @mock.patch('doppelganger.scripts.download_allocate_generate.os.path.exists')
    @mock.patch('doppelganger.scripts.download_allocate_generate.fetch_pums_data')
    def test_download_and_load_pums_data_dont_download(self, mock_fetch_pums_data, mock_exists,
                                                       mock_CleanedData, mock_pandas_to_csv):
        '''Verify:
        1. database is not called when files exist already
        2. files are written out properly
        '''
        configuration = self._mock_config()
        mock_exists.return_value = True

        download_allocate_generate.download_and_load_pums_data(
            output_dir=self._mock_params['output_dir'],
            state_id=self._mock_params['state_id'],
            puma_id=self._mock_params['puma_id'],
            configuration=configuration,
            db_host=self._mock_params['db_host'],
            db_database=self._mock_params['db_database'],
            db_schema=self._mock_params['db_schema'],
            db_user=self._mock_params['db_user'],
            db_password=self._mock_params['db_password']
        )

        # Check files are being written out properly
        households_filepath = os.path.sep.join([self._mock_params['output_dir'],
                                               download_allocate_generate.FILE_PATTERN.format(
                        self._mock_params['state_id'],
                        self._mock_params['puma_id'],
                        'households_pums.csv'
                    )
                ])
        persons_filepath = os.path.sep.join([self._mock_params['output_dir'],
                                            download_allocate_generate.FILE_PATTERN.format(
                        self._mock_params['state_id'],
                        self._mock_params['puma_id'],
                        'persons_pums.csv'
                    )
                ])
        self.assertFalse(mock_fetch_pums_data.called)
        self.assertTrue(mock_CleanedData.from_csv.called)
        mock_CleanedData.from_csv.assert_any_call(households_filepath)
        mock_CleanedData.from_csv.assert_any_call(persons_filepath)

    @mock.patch('doppelganger.scripts.download_allocate_generate.SegmentedData')
    @mock.patch('doppelganger.scripts.download_allocate_generate.BayesianNetworkModel')
    def test_create_bayes_net_segmenter(self, mock_BayesianNetworkModel, mock_SegmentedData):
        configuration = self._mock_config()
        mock_persons_data = mock.Mock()
        mock_households_data = mock.Mock()
        download_allocate_generate.create_bayes_net(
                output_dir=self._mock_params['output_dir'],
                state_id=self._mock_params['state_id'],
                puma_id=self._mock_params['puma_id'],
                configuration=configuration,
                persons_data=mock_persons_data,
                households_data=mock_households_data,
                person_segmenter=download_allocate_generate.person_segmenter,
                household_segmenter=download_allocate_generate.household_segmenter
            )
        mock_SegmentedData.from_data.assert_called()

        mock_SegmentedData.from_data.assert_any_call(
                cleaned_data=mock_persons_data,
                fields=list(configuration.person_fields),
                weight_field=inputs.PERSON_WEIGHT.name,
                segmenter=download_allocate_generate.person_segmenter
                )
        mock_SegmentedData.from_data.assert_any_call(
                cleaned_data=mock_households_data,
                fields=list(configuration.household_fields),
                weight_field=inputs.HOUSEHOLD_WEIGHT.name,
                segmenter=download_allocate_generate.household_segmenter
                )

    @mock.patch('doppelganger.scripts.download_allocate_generate.SegmentedData')
    @mock.patch('doppelganger.scripts.download_allocate_generate.BayesianNetworkModel')
    def test_create_bayes_net_model(self, mock_BayesianNetworkModel, mock_SegmentedData):
        configuration = self._mock_config()
        mock_persons_data = mock.Mock()
        mock_households_data = mock.Mock()
        mock_training_data = mock.Mock()
        mock_SegmentedData.from_data.return_value = mock_training_data

        download_allocate_generate.create_bayes_net(
                output_dir=self._mock_params['output_dir'],
                state_id=self._mock_params['state_id'],
                puma_id=self._mock_params['puma_id'],
                configuration=configuration,
                persons_data=mock_persons_data,
                households_data=mock_households_data,
                person_segmenter=download_allocate_generate.person_segmenter,
                household_segmenter=download_allocate_generate.household_segmenter
            )

        mock_BayesianNetworkModel.train.assert_any_call(
                    input_data=mock_training_data,
                    structure=configuration.person_structure,
                    fields=configuration.person_fields
                )

    @mock.patch('doppelganger.scripts.download_allocate_generate.HouseholdAllocator')
    @mock.patch('doppelganger.scripts.download_allocate_generate.Marginals')
    def test_download_tracts_data_dont_download(self, mock_Marginals, mock_HouseholdAllocator):
        mock_Marginals.from_csv.return_value = True
        mock_households_data = mock.Mock()
        mock_persons_data = mock.Mock()

        download_allocate_generate.download_tract_data(
            output_dir=self._mock_params['output_dir'],
            state_id=self._mock_params['state_id'],
            puma_id=self._mock_params['puma_id'],
            census_api_key=self._mock_params['census_api_key'],
            puma_tract_mappings=self._mock_params['puma_tract_mappings'],
            households_data=mock_households_data,
            persons_data=mock_persons_data,
        )
        self.assertFalse(mock_Marginals.from_census_data.called)

    @mock.patch('builtins.open')  # python 3 compatible
    @mock.patch('doppelganger.scripts.download_allocate_generate.HouseholdAllocator')
    @mock.patch('doppelganger.scripts.download_allocate_generate.Marginals')
    def test_download_tracts_data_download(self, mock_Marginals, mock_HouseholdAllocator,
                                           mock_open):
        mock_households_data = mock.Mock()
        mock_persons_data = mock.Mock()
        mock_returned_marginals = mock.Mock()
        mock_returned_marginals.data = [1, 2, 3]  # To pass len<=1 check
        mock_Marginals.from_census_data.return_value = mock_returned_marginals
        mock_Marginals.from_csv.side_effect = [  # File only found after downloading
                    Exception('File not found'),
                    mock_returned_marginals
                ]

        download_allocate_generate.download_tract_data(
            output_dir=self._mock_params['output_dir'],
            state_id=self._mock_params['state_id'],
            puma_id=self._mock_params['puma_id'],
            census_api_key=self._mock_params['census_api_key'],
            puma_tract_mappings=self._mock_params['puma_tract_mappings'],
            households_data=mock_households_data,
            persons_data=mock_persons_data,
        )
        self.assertTrue(mock_Marginals.from_census_data.called)
        mock_HouseholdAllocator.from_cleaned_data.assert_any_call(
                    marginals=mock_returned_marginals,
                    households_data=mock_households_data,
                    persons_data=mock_persons_data
                )

    @mock.patch('doppelganger.scripts.download_allocate_generate.Population')
    def test_generate_synths(self, mock_Population):
        mock_person_model = mock.Mock()
        mock_household_model = mock.Mock()
        mock_allocator = mock.Mock()
        download_allocate_generate.generate_synthetic_people_and_households(
                output_dir=self._mock_params['output_dir'],
                state_id=self._mock_params['state_id'],
                puma_id=self._mock_params['puma_id'],
                allocator=mock_allocator,
                person_model=mock_person_model,
                household_model=mock_household_model
            )
        mock_Population.generate.assert_any_call(
                    household_allocator=mock_allocator,
                    person_model=mock_person_model,
                    household_model=mock_household_model
                )



# Copyright 2017 Sidewalk Labs | https://www.apache.org/licenses/LICENSE-2.0

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import unittest

from doppelganger import inputs


class InputsTest(unittest.TestCase):

    def test_yyyy_to_age(self):
        for sample_birthday in [19490301, '19490301', '194903', 194903]:
            age = inputs.yyyy_to_age(sample_birthday)
            self.assertEqual(age, 68)

    def test_yyyy_to_age_none(self):
        for sample_birthday in ['', float('nan')]:
            age = inputs.yyyy_to_age(sample_birthday)
            self.assertEqual(age, inputs.UNKNOWN)

    def test_age_discrete(self):
        self.assertEquals(inputs.age_discrete(''), None)
        self.assertEquals(inputs.age_discrete('12'), '0-17')
        self.assertEquals(inputs.age_discrete('20'), '18-34')
        self.assertEquals(inputs.age_discrete('35'), '35-64')
        self.assertEquals(inputs.age_discrete('65'), '65+')

    def test_num_people_discrete(self):
        self.assertEquals(inputs.num_people_discrete('5'), '4+')

    def test_work_status(self):
        self.assertEquals(inputs.work_status(''), 'under16')
        self.assertEquals(inputs.work_status('1'), 'working')
        self.assertEquals(inputs.work_status('2'), 'working')
        self.assertEquals(inputs.work_status('4'), 'working')
        self.assertEquals(inputs.work_status('5'), 'working')
        self.assertEquals(inputs.work_status('3'), 'not_working')
        self.assertEquals(inputs.work_status('6'), 'not_working')

    def test_educational_attainment(self):
        self.assertEquals(inputs.educational_attainment(''), 'under3')
        self.assertEquals(inputs.educational_attainment('bb'), 'under3')

    def test_gender_named(self):
        self.assertEquals(inputs.gender_named(''), None)
        self.assertEquals(inputs.gender_named('3'), None)
        self.assertEquals(inputs.gender_named('1'), 'M')
        self.assertEquals(inputs.gender_named('2'), 'F')
